<?xml version="1.0"?>
<!--
Copyright (C) 2002, 2003  Scott Lystig Fritchie
Copyright (C) 2006  Romain Lenglet
All rights reserved.
The contents of this file are subject to the Modified BSD License.
See the file "COPYING" for licensing details.
-->

<xsl:stylesheet version='1.0'
    xmlns:xsl='http://www.w3.org/1999/XSL/Transform'
    xmlns:dryverl='http://objectweb.org/dryverl-0.1'>

<xsl:output method='text'/>

<!-- Empty template in case there is no gen-driver element! -->
<xsl:template match='/dryverl:def-bindings[not(dryverl:gen-driver)]'>
</xsl:template>

<xsl:template match='/dryverl:def-bindings[dryverl:gen-driver]'>
<xsl:variable name='prefix' select='dryverl:gen-driver/@c-prefix'/>
<xsl:variable name='constant-prefix'
	      select='dryverl:gen-driver/@c-constant-prefix'/>
#ifndef	__<xsl:value-of select='$constant-prefix'/>_H
#define	__<xsl:value-of select='$constant-prefix'/>_H

#include &lt;pthread.h&gt;
#include &lt;erl_driver.h&gt;
#include &lt;ei.h&gt;

#define <xsl:value-of select='$constant-prefix'/>_COMMAND_GET_OPTION 0
#define <xsl:value-of select='$constant-prefix'/>_COMMAND_SET_OPTION 1
<xsl:variable name='first-user-command-id' select='2'/>

<xsl:for-each select='dryverl:def-erlang-to-c-binding'>
#define <xsl:value-of select='$constant-prefix'/>_COMMAND_<xsl:value-of select='@name'/><xsl:text> </xsl:text><xsl:value-of select='position() + $first-user-command-id - 1'/>
</xsl:for-each>

#define <xsl:value-of select='$constant-prefix'/>_OPTION_ASYNC 0

<xsl:if test='dryverl:def-value-map'>
#define <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAG_PINNED 1
#define <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAG_ALLOCED 2
#define <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAGS_PUBLIC_MASK 1

typedef enum {
    FIND_ID,
    FIND_VALUE,
    FIND_VALUE_AND_ADD_IF_NOT_FOUND,
    REMOVE_ID,
    REMOVE_VALUE,
    DESTROY_ID,
    DESTROY_VALUE
} <xsl:value-of select='$prefix'/>_value_map_entry_op_t;
</xsl:if>

<!-- Free user code -->
<xsl:apply-templates select='dryverl:gen-driver/dryverl:def-h-file-header'/>

<xsl:for-each select='dryverl:def-value-map'>
<xsl:variable name='value-map-t'><xsl:value-of select='$prefix'/>_value_map_<xsl:value-of select='@name'/>_t</xsl:variable>
typedef struct <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/> {
    unsigned long id;
<xsl:text>    </xsl:text><xsl:value-of select='@c-type'/> value;
    unsigned char flags;
    struct <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/> *previous;
    struct <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/> *next;
<xsl:if test='dryverl:def-id-hash-map'>
    struct <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/> *previous_same_id_hash;
    struct <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/> *next_same_id_hash;
</xsl:if>
<xsl:if test='dryverl:def-value-hash-map'>
    struct <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/> *previous_same_value_hash;
    struct <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/> *next_same_value_hash;
</xsl:if>
} <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t;

typedef struct <xsl:value-of select='$prefix'/>_value_map_<xsl:value-of select='@name'/> {
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t entries[<xsl:value-of select='@pre-allocated-size'/>];
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t *first_used_entry;
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t *first_unused_entry;
    unsigned long next_id;
    pthread_mutex_t mutex;
<xsl:if test='dryverl:def-id-hash-map'>
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t *id_hash_table[<xsl:value-of select='dryverl:def-id-hash-map/@table-size'/>];
</xsl:if>
<xsl:if test='dryverl:def-value-hash-map'>
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t *value_hash_table[<xsl:value-of select='dryverl:def-value-hash-map/@table-size'/>];
</xsl:if>
} <xsl:value-of select='$value-map-t'/>;

extern int <xsl:value-of select='$prefix'/>_value_map_find_entry_<xsl:value-of select='@name'/>(<xsl:value-of select='$value-map-t'/> *, <xsl:value-of select='$prefix'/>_value_map_entry_op_t, unsigned long *, <xsl:value-of select='@c-type'/> *, unsigned char *, unsigned char, unsigned char);
</xsl:for-each>

typedef struct <xsl:value-of select='$prefix'/>_driver_data {
    ErlDrvPort port;
    int option_async;
<xsl:for-each select='dryverl:def-value-map'>
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_value_map_<xsl:value-of select='@name'/>_t value_map_<xsl:value-of select='@name'/>;
</xsl:for-each>
} <xsl:value-of select='$prefix'/>_driver_data_t;

<xsl:for-each select='dryverl:def-erlang-to-c-binding'>
<xsl:variable name='has-call-vars' select='boolean(dryverl:def-c-call-variable)'/>
<xsl:variable name='has-output' select='boolean(dryverl:encode-output)'/>

<xsl:if test='$has-call-vars'>
typedef struct <xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_variables {
<xsl:for-each select='dryverl:def-c-call-variable'>
<xsl:text>    </xsl:text><xsl:value-of select='@c-type'/><xsl:text> </xsl:text><xsl:value-of select='@name'/>;
</xsl:for-each>} <xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_variables_t;
</xsl:if>

typedef struct <xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_state {
    <xsl:value-of select='$prefix'/>_driver_data_t *driver_data;
    int async_call;
<xsl:if test='$has-output'>
    erlang_pid reply_pid;
    erlang_ref reply_tag;
    <!-- If not NULL, used to return the output term buffer, in case the call is
         performed using port_call and not port_command -->
    <!-- WARNING: *output_main_term_buffer and *output_status must not be
         modified if async_call is true, or if NULL -->
    char **output_main_term_buffer;
    int output_main_term_buffer_size;
    int *output_status;
</xsl:if>
<xsl:if test='$has-call-vars'>
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_variables_t call_variables;
</xsl:if>
} <xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_state_t;
</xsl:for-each>

#endif	/* __<xsl:value-of select='$constant-prefix'/>_H */

</xsl:template>

</xsl:stylesheet>

