<?xml version="1.0"?>
<!--
Copyright (C) 2002, 2003  Scott Lystig Fritchie
Copyright (C) 2006  Romain Lenglet
All rights reserved.
The contents of this file are subject to the Modified BSD License.
See the file "COPYING" for licensing details.
-->

<xsl:stylesheet version='1.0'
    xmlns:xsl='http://www.w3.org/1999/XSL/Transform'
    xmlns:dryverl='http://objectweb.org/dryverl-0.1'>

<xsl:output method='text'/>

<xsl:variable name='runtime-checks'
    select='not(/dryverl:def-bindings/dryverl:gen-driver/@disable-runtime-checks)'/>
<xsl:variable name='prefix' select='/dryverl:def-bindings/dryverl:gen-driver/@c-prefix'/>
<xsl:variable name='constant-prefix' select='/dryverl:def-bindings/dryverl:gen-driver/@c-constant-prefix'/>
<xsl:variable name='driver-data-t'><xsl:value-of select='$prefix'/>_driver_data_t</xsl:variable>

<!-- Empty template in case there is no gen-driver element! -->
<xsl:template match='/dryverl:def-bindings[not(dryverl:gen-driver)]'>
</xsl:template>

<xsl:template match='/dryverl:def-bindings[dryverl:gen-driver]'>
#include "<xsl:value-of select='dryverl:gen-driver/@c-file-name'/>.h"

#include &lt;string.h&gt;
#include &lt;unistd.h&gt;
#include &lt;limits.h&gt;

static ErlDrvTermData <xsl:value-of select='$prefix'/>_atom_drvreply;

static int
<xsl:value-of select='$prefix'/>_init(void)
{
    <xsl:value-of select='$prefix'/>_atom_drvreply = driver_mk_atom("drvreply");
    {
    <xsl:apply-templates select='dryverl:gen-driver/dryverl:def-c-driver-init'/>
    }
    return 0;
}

static ErlDrvData
<xsl:value-of select='$prefix'/>_start(ErlDrvPort port, char *command)
{
    <xsl:value-of select='$driver-data-t'/> *driver_data = NULL;
<xsl:if test='dryverl:def-value-map'>
    int i = 0;
</xsl:if>

    <xsl:if test='$runtime-checks'>
    if (!port) goto error;
    </xsl:if>

    driver_data = (<xsl:value-of select='$driver-data-t'/> *) driver_alloc(sizeof(<xsl:value-of select='$driver-data-t'/>));
    <xsl:if test='$runtime-checks'>
    if (! driver_data) goto error;
    </xsl:if>

    memset(driver_data, 0, sizeof(<xsl:value-of select='$driver-data-t'/>));
    driver_data->port = port;

<xsl:for-each select='dryverl:def-value-map'>
    {
        <xsl:value-of select='$prefix'/>_value_map_<xsl:value-of select='@name'/>_t *value_map =
            &amp;driver_data->value_map_<xsl:value-of select='@name'/>;
<xsl:if test='@pre-allocated-size &lt; 2'>
<xsl:message terminate="yes">
The @pre-allocated-size attribute of the &lt;value-map name="<xsl:value-of select='@name'/>" .../> element must be an integer value >= 2.
</xsl:message>
</xsl:if>
        value_map->first_unused_entry = &amp;value_map->entries[0];
        value_map->entries[0].next = &amp;value_map->entries[1];
        for (i = 1; i &lt; <xsl:value-of select='@pre-allocated-size - 1'/>; i++) {
            value_map->entries[i].previous = &amp;value_map->entries[i-1];
            value_map->entries[i].next = &amp;value_map->entries[i+1];
        }
        value_map->entries[<xsl:value-of select='@pre-allocated-size - 1'/>].previous = &amp;value_map->entries[<xsl:value-of select='@pre-allocated-size - 2'/>];
        pthread_mutex_init(&amp;value_map->mutex, NULL);
    }

</xsl:for-each>

    return (ErlDrvData) driver_data;

<xsl:if test='$runtime-checks'>
error:
    if (driver_data) driver_free(driver_data);
    return ERL_DRV_ERROR_GENERAL;
</xsl:if>
}

static void
<xsl:value-of select='$prefix'/>_stop(ErlDrvData drv_data)
{
    <xsl:value-of select='$driver-data-t'/> *driver_data = (<xsl:value-of select='$driver-data-t'/> *) drv_data;

<xsl:if test='dryverl:def-value-map'>
    int all_pinned_entries = 1;

    while (all_pinned_entries) {

        all_pinned_entries = 0;

<xsl:for-each select='dryverl:def-value-map'>
        {
        <!-- Remove all used entries which are not pinned -->
        int has_unpinned_entries = 1;
        int pinned_entries = 0;
        while (has_unpinned_entries) {
            <xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t *entry;
            has_unpinned_entries = 0;
            pinned_entries = 0;
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            pthread_mutex_lock(&amp;driver_data->value_map_<xsl:value-of select='@name'/>.mutex)
            <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
            entry = driver_data->value_map_<xsl:value-of select='@name'/>.first_used_entry;
            while (entry &amp;&amp; (! (entry->flags &amp; <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAG_PINNED))) {
                pinned_entries++;
                entry = entry->next;
            }
            if (entry) {
                has_unpinned_entries = 1;
                <xsl:value-of select='$prefix'/>_value_map_find_entry_<xsl:value-of select='@name'/>(
                    &amp;driver_data->value_map_<xsl:value-of select='@name'/>,
		    DESTROY_ID, &amp;entry->id, NULL, NULL, 0, 0);
            }
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            pthread_mutex_unlock(&amp;driver_data->value_map_<xsl:value-of select='@name'/>.mutex)
            <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
        }
        all_pinned_entries += pinned_entries;
        }
</xsl:for-each>

        if (all_pinned_entries) {
            sleep(1);
        }
    }

error:
</xsl:if>

    driver_free(driver_data);

}

static int
<xsl:value-of select='$prefix'/>_call_get_option(<xsl:value-of select='$driver-data-t'/> *driver_data, char *buf, int orig_index,
                   int len, char **rbuf, int rlen, unsigned int *flags)
{
    int index = orig_index;
    unsigned long option_id = 0L;

    <xsl:if test='$runtime-checks'>
    if ((!buf) || (len &lt; 0) || (!rbuf) || (!*rbuf) || (rlen &lt;= 0))
        goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_ulong(buf, &amp;index, &amp;option_id)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > len) goto error;
    </xsl:if>

    switch (option_id) {
    case <xsl:value-of select='$constant-prefix'/>_OPTION_ASYNC:
        {
            index = 0;
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            ei_encode_version(NULL, &amp;index)
            <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            ei_encode_boolean(NULL, &amp;index, driver_data->option_async)
            <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
            if (index > rlen) {
                *rbuf = (char*) driver_alloc(index);
                <xsl:if test='$runtime-checks'>
                if (! *rbuf) goto error;
                </xsl:if>
            }
            index = 0;
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            ei_encode_version(*rbuf, &amp;index)
            <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            ei_encode_boolean(*rbuf, &amp;index, driver_data->option_async)
            <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
        }
        break;
<xsl:if test='$runtime-checks'>
    default:
        goto error;
        break;
</xsl:if>
    }

    return index;

<xsl:if test='$runtime-checks'>
error:
    return (int) ERL_DRV_ERROR_GENERAL;
</xsl:if>

}

static int
<xsl:value-of select='$prefix'/>_call_set_option(<xsl:value-of select='$driver-data-t'/> *driver_data, char *buf, int orig_index,
                   int len, char **rbuf, int rlen, unsigned int *flags)
{
    int index = orig_index;
    int tuple_arity = 0;
    unsigned long option_id = 0L;

    <xsl:if test='$runtime-checks'>
    if ((!buf) || (len &lt; 0) || (!rbuf) || (!*rbuf) || (rlen &lt;= 0))
        goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_tuple_header(buf, &amp;index, &amp;tuple_arity)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > len) goto error;
    </xsl:if>
    <xsl:if test='$runtime-checks'>
    if (tuple_arity != 2) goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_ulong(buf, &amp;index, &amp;option_id)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > len) goto error;
    </xsl:if>

    switch (option_id) {
    case <xsl:value-of select='$constant-prefix'/>_OPTION_ASYNC:
        {
            int async = 0;
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            ei_decode_boolean(buf, &amp;index, &amp;async)
            <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
            <xsl:if test='$runtime-checks'>
            if (index > len) goto error;
            </xsl:if>
            driver_data->option_async = async;
        }
        break;
<xsl:if test='$runtime-checks'>
    default:
        goto error;
        break;
</xsl:if>
    }

    index = 0;
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_version((char*)0, &amp;index)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_atom((char*)0, &amp;index, "ok")
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    if (index > rlen) {
        *rbuf = (char*) driver_alloc(index);
        <xsl:if test='$runtime-checks'>
        if (! *rbuf) goto error;
        </xsl:if>
    }
    index = 0;
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_version(*rbuf, &amp;index)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_atom(*rbuf, &amp;index, "ok")
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;

    return index;

<xsl:if test='$runtime-checks'>
error:
    return (int) ERL_DRV_ERROR_GENERAL;
</xsl:if>

}

<xsl:for-each select='dryverl:def-erlang-to-c-binding'>
<xsl:variable name='binding' select='.'/>
<xsl:variable name='call-state-t'><xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_state_t</xsl:variable>
<xsl:variable name='has-call-vars' select='boolean(dryverl:def-c-call-variable)'/>
<xsl:variable name='has-output' select='boolean(dryverl:encode-output)'/>
<xsl:variable name='has-output-byref-bins' select='boolean(dryverl:encode-output/descendant::dryverl:encode-output-byref-binary)'/>

<xsl:if test='$has-output'>
static void
<xsl:value-of select='$prefix'/>_encode_output_<xsl:value-of select='@name'/>(<xsl:value-of select='$call-state-t'/> *call_state)
{
    int can_reply_directly_to_port_call = 0;
    ErlDrvBinary *main_term_binary = NULL;
    char *main_term_buffer = NULL;
    char *temp_buffer = NULL;
    int main_term_buffer_index = 0;
    int main_term_buffer_size = 0;
    int old_buffer_size = 0;
    int main_term_buffer_was_alloced = 0;
    int temp_index;
    int allowed_terms_count = 1;

    ErlDrvTermData *driver_term = NULL;
    int driver_term_size = 0;
    <!-- NOTE: 10 is the index for encoding the first byref binary in the
         driver_term -->
    <xsl:if test='$has-output-byref-bins'>
    int driver_term_index = 10;
    int byref_binaries_count = 0;
    </xsl:if>

<xsl:if test='$has-call-vars'>
    <xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_variables_t *call_variables = &amp;call_state->call_variables;
</xsl:if>
<xsl:for-each select='dryverl:encode-output/dryverl:def-c-local-variable'>
<xsl:text>    </xsl:text><xsl:value-of select='@c-type'/> local_variable_<xsl:value-of select='@name'/><xsl:if test='* or text()'> = (<xsl:value-of select='text()'/>)</xsl:if>;
</xsl:for-each>

    <xsl:if test='$runtime-checks'>
    if (! call_state) goto argument_error;
    </xsl:if>

    can_reply_directly_to_port_call = (! call_state->async_call)
        &amp;&amp; (call_state->output_main_term_buffer != NULL);

    if (can_reply_directly_to_port_call) {
        main_term_buffer = *(call_state->output_main_term_buffer);
        main_term_buffer_size = call_state->output_main_term_buffer_size;
    } else {
        main_term_buffer_size = 128;
        main_term_binary = driver_alloc_binary(main_term_buffer_size);
        <xsl:if test='$runtime-checks'>
        if (! main_term_binary) goto memory_error;
        </xsl:if>
        main_term_buffer = main_term_binary->orig_bytes;
    }

    <!-- Encode the common part of the replied term which contains the main
         term binary -->
    <!-- NOTE: We assume here that the buffer is large enough to encode the
         common part. This is the case with buffers provided by port_call (128
         octets), and if I use a binary allocated above (128 octets also). -->
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_version(main_term_buffer, &amp;main_term_buffer_index)
    <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto memory_error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_tuple_header(main_term_buffer, &amp;main_term_buffer_index, 3)
    <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto memory_error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_pid(main_term_buffer, &amp;main_term_buffer_index,
        &amp;call_state->reply_pid)
    <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto memory_error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_encode_ref(main_term_buffer, &amp;main_term_buffer_index,
        &amp;call_state->reply_tag)
    <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto memory_error;
    </xsl:if>


    <!-- Encode the last member of the tuple (the actual returned term!), and
         get the references to binaries to be returned by reference, if any -->
    <xsl:apply-templates select='dryverl:encode-output/dryverl:encode-output-main-term-and-byref-binaries'/>

    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count != 0) goto encoding_error;
    </xsl:if>


    if (can_reply_directly_to_port_call) {
        if (main_term_binary) {
            <!-- There are binaries to return by reference, return the
                 'noreply' atom to the port_call call, and send the term and
                 binaries using driver_output_term below -->
            *(call_state->output_status) = 0;
        } else {
            <!-- There was no binary to return by reference, return the main
                 term to the port_call call -->
            <xsl:if test='$runtime-checks'>
            if (main_term_buffer_index &lt;= 0) goto encoding_error;
            </xsl:if>
            *(call_state->output_main_term_buffer) = main_term_buffer;
            *(call_state->output_status) = main_term_buffer_index;
            goto deallocate;
        }
    }


    if (! driver_term) {
        <xsl:choose>
        <xsl:when test='$has-output-byref-bins'>
        driver_term_size = 13;
        </xsl:when>
        <xsl:otherwise>
        driver_term_size = 12;
        </xsl:otherwise>
        </xsl:choose>
        driver_term = (ErlDrvTermData*) driver_alloc(
            sizeof(ErlDrvTermData) * driver_term_size);
        <xsl:if test='$runtime-checks'>
        if (! driver_term) goto memory_error;
        </xsl:if>
    }

    driver_term[0] = ERL_DRV_ATOM;
    driver_term[1] =     <xsl:value-of select='$prefix'/>_atom_drvreply;
    driver_term[2] = ERL_DRV_PORT;
    driver_term[3] =     driver_mk_port(call_state->driver_data->port);
    driver_term[4] = ERL_DRV_INT;
    driver_term[5] =     <xsl:value-of select='$constant-prefix'/>_COMMAND_<xsl:value-of select='@name'/>;
    driver_term[6] = ERL_DRV_BINARY;
    driver_term[7] =     (ErlDrvTermData) main_term_binary;
    driver_term[8] =     main_term_buffer_index;
    driver_term[9] =     0;

    <xsl:choose>
    <xsl:when test='$has-output-byref-bins'>
    driver_term[driver_term_index++] = ERL_DRV_NIL;
    if (byref_binaries_count > 0) {
        driver_term[driver_term_index++] = ERL_DRV_LIST;
        driver_term[driver_term_index++] =     byref_binaries_count+1;
    }
    driver_term[driver_term_index++] = ERL_DRV_TUPLE;
    driver_term[driver_term_index++] =     5;
    driver_output_term(call_state->driver_data->port,
        driver_term, driver_term_index);
    </xsl:when>
    <xsl:otherwise>
    driver_term[10] = ERL_DRV_TUPLE;
    driver_term[11] =     4;
    driver_output_term(call_state->driver_data->port,
        driver_term, 12);
    </xsl:otherwise>
    </xsl:choose>

deallocate:
    if (driver_term)
        driver_free(driver_term);
    driver_free(call_state);
    return;

<xsl:if test='$runtime-checks'>
argument_error:
    if (can_reply_directly_to_port_call)
        *(call_state->output_status) = (int) ERL_DRV_ERROR_GENERAL;
    else
        driver_failure_atom(call_state->driver_data->port, "badarg");
    goto error_deallocate;

memory_error:
    if (can_reply_directly_to_port_call)
        *(call_state->output_status) = (int) ERL_DRV_ERROR_GENERAL;
    else
        driver_failure_atom(call_state->driver_data->port,
            "memory_allocation_error");
    goto error_deallocate;

encoding_error:
    if (can_reply_directly_to_port_call)
        *(call_state->output_status) = (int) ERL_DRV_ERROR_GENERAL;
    else
        driver_failure_atom(call_state->driver_data->port,
            "output_encoding_error");
    goto error_deallocate;

error_deallocate:
    if (main_term_buffer_was_alloced)
        driver_free(main_term_buffer);
    if (main_term_binary)
        driver_free_binary(main_term_binary);
    goto deallocate;
</xsl:if>

}
</xsl:if> <!-- test='$has-output' -->

static void
<xsl:value-of select='$prefix'/>_call_<xsl:value-of select='@name'/>(<xsl:value-of select='$call-state-t'/> *call_state)
{
<xsl:if test='$has-call-vars'>
    <xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_variables_t *call_variables = &amp;call_state->call_variables;
</xsl:if>
<xsl:for-each select='dryverl:execute-body/dryverl:def-c-local-variable'>
<xsl:text>    </xsl:text><xsl:value-of select='@c-type'/> local_variable_<xsl:value-of select='@name'/><xsl:if test='* or text()'> = (<xsl:value-of select='text()'/>)</xsl:if>;
</xsl:for-each>

    <xsl:apply-templates select='dryverl:execute-body/dryverl:process-c-call-variables'/>

<xsl:if test='$has-output'>
    <xsl:value-of select='$prefix'/>_encode_output_<xsl:value-of select='@name'/>(call_state);
</xsl:if>
}

static void
<xsl:value-of select='$prefix'/>_decode_input_<xsl:value-of select='@name'/>(
    <xsl:value-of select='$driver-data-t'/> *driver_data,
        int input_main_term_binary_buffer_index,
        int input_main_term_binary_buffer_size,
    <xsl:if test='dryverl:encode-input/dryverl:encode-input-byref-binary-list'>
    <!-- These parameters are != 0 only when the caller is ..._outputv -->
    ErlIOVec *ev, int input_main_term_binary_vector_index,
        int input_byref_binaries_sizes_list_binary_size,
    </xsl:if>
    <!-- These parameters are != 0 only when the caller is ..._call -->
    char *input_main_term_binary_buffer
        <xsl:if test='dryverl:encode-output'>, char **output_main_term_buffer,
        int output_main_term_buffer_size, int *output_status</xsl:if>)
{
    <xsl:value-of select='$call-state-t'/> *call_state = NULL;
<xsl:if test='$runtime-checks'>
    int allowed_terms_count = 0;
</xsl:if>
    char *main_term_buffer = (char*) NULL;
    int main_term_buffer_index = input_main_term_binary_buffer_index;
    int main_term_buffer_size = input_main_term_binary_buffer_size;
<xsl:if test='dryverl:encode-input/dryverl:encode-input-byref-binary-list'>
    int version = 0;
    char *byref_binaries_sizes_list_buffer = (char*) NULL;
    int byref_binaries_sizes_list_buffer_index = 0;
    int byref_binaries_sizes_list_buffer_size = 0;
    int byref_binaries_sizes_list_is_a_string = 0;

    int byref_binaries_count = 0;
<xsl:if test='$runtime-checks'>
    unsigned long allowed_byref_binaries_count;
</xsl:if>
    int byref_binary_vector_index = 0;
    int byref_binary_buffer_index = 0;

    unsigned long temp_byref_binary_size = 0;
</xsl:if>
<xsl:if test='$has-call-vars'>
    <xsl:value-of select='$prefix'/>_<xsl:value-of select='@name'/>_call_variables_t *call_variables;
</xsl:if>
<xsl:for-each select='dryverl:decode-input/dryverl:def-c-local-variable'>
<xsl:text>    </xsl:text><xsl:value-of select='@c-type'/> local_variable_<xsl:value-of select='@name'/><xsl:if test='* or text()'> = (<xsl:value-of select='text()'/>)</xsl:if>;
</xsl:for-each>

    <xsl:if test='dryverl:encode-input/dryverl:encode-input-byref-binary-list'>
    if (ev) {
        <xsl:if test='$runtime-checks'>
        if (input_main_term_binary_vector_index >= ev->vsize) goto argument_error;
        </xsl:if>
        main_term_buffer = ev->iov[input_main_term_binary_vector_index].iov_base;
        <xsl:if test='$runtime-checks'>
        if (main_term_buffer_size >
                ev->iov[input_main_term_binary_vector_index].iov_len) goto argument_error;
        </xsl:if>

        <!-- Find the pointer to the byref binaries sizes' list binary -->
        if (main_term_buffer_size ==
                ev->iov[input_main_term_binary_vector_index].iov_len) {
            byref_binary_vector_index = input_main_term_binary_vector_index + 1;
            byref_binaries_sizes_list_buffer_index = 0;
            <xsl:if test='$runtime-checks'>
            if (byref_binary_vector_index >= ev->vsize) goto argument_error;
            </xsl:if>
        } else {
            byref_binary_vector_index = input_main_term_binary_vector_index;
            byref_binaries_sizes_list_buffer_index = main_term_buffer_size;
        }
        byref_binaries_sizes_list_buffer_size =
            byref_binaries_sizes_list_buffer_index +
            input_byref_binaries_sizes_list_binary_size;
        <xsl:if test='$runtime-checks'>
        if (byref_binaries_sizes_list_buffer_size >
                ev->iov[byref_binary_vector_index].iov_len)
            goto argument_error;
        </xsl:if>
        byref_binaries_sizes_list_buffer =
            ev->iov[byref_binary_vector_index].iov_base;

        <!-- Decode the start of the byref binaries sizes' list binary -->
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_decode_version(byref_binaries_sizes_list_buffer,
            &amp;byref_binaries_sizes_list_buffer_index, &amp;version)
        <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
        <xsl:if test='$runtime-checks'>
        if (byref_binaries_sizes_list_buffer_index
                > byref_binaries_sizes_list_buffer_size) goto decoding_error;
        </xsl:if>

	<!-- The sizes list may be encoded as a string, if all sizes are < 256.
             Decode the header appropriately. -->
        <xsl:if test='$runtime-checks'>if (</xsl:if>
	ei_get_type(byref_binaries_sizes_list_buffer,
            &amp;byref_binaries_sizes_list_buffer_index,
	    &amp;byref_binaries_sizes_list_is_a_string,
	    &amp;byref_binaries_count)
        <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
	<xsl:if test='$runtime-checks'>
        if (byref_binaries_count &lt;= 0) goto argument_error;
        </xsl:if>

	if (byref_binaries_sizes_list_is_a_string =
 	       (byref_binaries_sizes_list_is_a_string == ERL_STRING_EXT)) {
	    <!-- Skip the string type and length. -->
            <xsl:if test='$runtime-checks'>
              if ((byref_binaries_sizes_list_buffer_index + 1 + 2)
                  > byref_binaries_sizes_list_buffer_size) goto decoding_error;
            </xsl:if>
            byref_binaries_sizes_list_buffer_index += 1 + 2;
        } else {
	    <!-- Skip the list type and length. -->
            <xsl:if test='$runtime-checks'>
              if ((byref_binaries_sizes_list_buffer_index + 1 + 4)
                  > byref_binaries_sizes_list_buffer_size) goto decoding_error;
            </xsl:if>
            <xsl:if test='$runtime-checks'>if (</xsl:if>
            ei_decode_list_header(byref_binaries_sizes_list_buffer,
                &amp;byref_binaries_sizes_list_buffer_index,
                NULL)
            <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
        }

        <!-- Find the pointer to the first byref binary -->
        if (byref_binaries_sizes_list_buffer_size ==
                ev->iov[byref_binary_vector_index].iov_len) {
            byref_binary_vector_index++;
            byref_binary_buffer_index = 0;
            <xsl:if test='$runtime-checks'>
            if (byref_binary_vector_index >= ev->vsize) goto argument_error;
            </xsl:if>
        } else {
            byref_binary_buffer_index =
                byref_binaries_sizes_list_buffer_size;
        }

    } else
    </xsl:if>
        main_term_buffer = input_main_term_binary_buffer;
    <xsl:if test='$runtime-checks'>
    if (! main_term_buffer) goto argument_error;
    </xsl:if>

    call_state = (<xsl:value-of select='$call-state-t'/>*)
        driver_alloc(sizeof(<xsl:value-of select='$call-state-t'/>));
    <xsl:if test='$runtime-checks'>
    if (! call_state) goto memory_error;
    </xsl:if>
    memset(call_state, 0, sizeof(<xsl:value-of select='$call-state-t'/>));

    <!-- Decode information to send back output, used only in case we use
         driver_output_term() to send it as a message to the port owner -->
<xsl:if test='dryverl:encode-output'>
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_pid(main_term_buffer, &amp;main_term_buffer_index,
        &amp;call_state->reply_pid)
    <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto decoding_error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_ref(main_term_buffer, &amp;main_term_buffer_index,
        &amp;call_state->reply_tag)
    <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto decoding_error;
    </xsl:if>
</xsl:if>

    call_state->driver_data = driver_data;
    call_state->async_call = 0;

<xsl:if test='dryverl:encode-output'>
    call_state->output_main_term_buffer = output_main_term_buffer;
    call_state->output_main_term_buffer_size = output_main_term_buffer_size;
    call_state->output_status = output_status;
</xsl:if>

<xsl:if test='$has-call-vars'>
    call_variables = &amp;call_state->call_variables;
<xsl:for-each select='dryverl:def-c-call-variable'>
    call_variables-><xsl:value-of select='@name'/> = (<xsl:apply-templates/>);
</xsl:for-each>
</xsl:if>

<xsl:if test='$runtime-checks'>
<xsl:if test='dryverl:encode-input/dryverl:encode-input-main-term'>
    allowed_terms_count = 1;
</xsl:if>
<xsl:if test='dryverl:encode-input/dryverl:encode-input-byref-binary-list'>
    allowed_byref_binaries_count = byref_binaries_count;
</xsl:if>
</xsl:if>
<xsl:apply-templates select='dryverl:decode-input/dryverl:assign-c-call-variables'/>

<xsl:if test='$runtime-checks'>
<xsl:if test='dryverl:encode-input/dryverl:encode-input-main-term'>
    if (allowed_terms_count != 0) goto decoding_error;
</xsl:if>
<xsl:if test='dryverl:encode-input/dryverl:encode-input-byref-binary-list'>
    if (allowed_byref_binaries_count != 0) goto decoding_error;
</xsl:if>
</xsl:if>

    if (driver_data->option_async) {
        call_state->async_call = 1;
        driver_async(driver_data->port, NULL,
            (void(*)(void*)) <xsl:value-of select='$prefix'/>_call_<xsl:value-of select='@name'/>,
            call_state, driver_free);
    } else {
        <xsl:value-of select='$prefix'/>_call_<xsl:value-of select='@name'/>(call_state);
    }

    return;

<xsl:if test='$runtime-checks'>
memory_error:
    driver_failure_atom(call_state->driver_data->port,
            "memory_allocation_error");
    return;

argument_error:
    driver_failure_atom(driver_data->port, "badarg");
    return;

decoding_error:
    driver_failure_atom(driver_data->port, "decoding_error");
    return;
</xsl:if>

}
</xsl:for-each>

static int 
<xsl:value-of select='$prefix'/>_call(ErlDrvData drv_data, unsigned int command, char *buf,
                   int len, char **rbuf, int rlen, unsigned int *flags)
{
    <xsl:value-of select='$driver-data-t'/> *driver_data = (<xsl:value-of select='$driver-data-t'/> *) drv_data;
    int index = 0;
    int version = 0;
<xsl:if test='dryverl:def-erlang-to-c-binding/dryverl:encode-output'>
    int tuple_arity = 0;
    int output_status = 0;
</xsl:if>

    <xsl:if test='$runtime-checks'>
    if ((!buf) || (len &lt; 0) || (!rbuf) || (!*rbuf) || (rlen &lt;= 0))
        goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_version(buf, &amp;index, &amp;version)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > len) goto error;
    </xsl:if>

    switch (command) {
    case <xsl:value-of select='$constant-prefix'/>_COMMAND_GET_OPTION:
        return <xsl:value-of select='$prefix'/>_call_get_option(
            driver_data, buf, index, len, rbuf, rlen, flags);
        break;
    case <xsl:value-of select='$constant-prefix'/>_COMMAND_SET_OPTION:
        return <xsl:value-of select='$prefix'/>_call_set_option(
            driver_data, buf, index, len, rbuf, rlen, flags);
        break;
<xsl:for-each select='dryverl:def-erlang-to-c-binding'>
    case <xsl:value-of select='$constant-prefix'/>_COMMAND_<xsl:value-of select='@name'/>:
        <xsl:if test='dryverl:encode-output'>
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_decode_tuple_header(buf, &amp;index, &amp;tuple_arity)
        <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
        <xsl:if test='$runtime-checks'>
        if (index > len) goto error;
        if (tuple_arity != 3) goto error;
        </xsl:if>
        </xsl:if>
        <xsl:value-of select='$prefix'/>_decode_input_<xsl:value-of select='@name'/>(
            driver_data, index, len,
            <xsl:if test='dryverl:encode-input/dryverl:encode-input-byref-binary-list'>
            NULL, 0, 0,
            </xsl:if>
            buf <xsl:if test='dryverl:encode-output'>, rbuf, rlen,
                &amp;output_status</xsl:if>);
        break;
</xsl:for-each>
    <xsl:if test='$runtime-checks'>
    default:
        goto error;
        break;
    </xsl:if>
    }

<xsl:if test='dryverl:def-erlang-to-c-binding/dryverl:encode-output'>
    if (output_status &lt; 0) {
        goto error;
    } else if (output_status > 0) {
        <!-- Output is sent as this function call's return -->
        return output_status;
    } else
        <!-- Output is sent using driver_output_term, not as this function
             call's return: in that case, return the 'noreply' atom here -->
</xsl:if>
    {
        index = 0;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_version(NULL, &amp;index)
        <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_atom(NULL, &amp;index, "noreply")
        <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
        if (index > rlen) {
            *rbuf = (char*) driver_alloc(index);
            <xsl:if test='$runtime-checks'>
            if (! *rbuf) goto error;
            </xsl:if>
        }
        index = 0;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_version(*rbuf, &amp;index)
        <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_atom(*rbuf, &amp;index, "noreply")
        <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
        return index;
    }

<xsl:if test='$runtime-checks'>
error:
    return (int) ERL_DRV_ERROR_GENERAL;
</xsl:if>

}

<xsl:if test='dryverl:def-erlang-to-c-binding[dryverl:encode-input/dryverl:encode-input-byref-binary-list]'>
static void
<xsl:value-of select='$prefix'/>_outputv(ErlDrvData drv_data, ErlIOVec *ev)
{
    <xsl:value-of select='$driver-data-t'/> *driver_data = (<xsl:value-of select='$driver-data-t'/> *) drv_data;
    int index = 0;
    int version = 0;
    unsigned long main_term_binary_size = 0L;
    unsigned long main_term_binary_buffer_size = 0L;
    unsigned long byref_binaries_sizes_list_binary_size = 0L;
    int main_term_binary_vector_start_index = 0;
    int main_term_binary_vector_index = 0;
    int tuple_arity = 0;
    unsigned long command_id;

    <xsl:if test='$runtime-checks'>
    if (!drv_data || !ev || (ev->vsize &lt; 2)) goto error;
    </xsl:if>

    <!-- Decode the header binary -->
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_version(ev->iov[1].iov_base, &amp;index, &amp;version)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > ev->iov[1].iov_len) goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_tuple_header(ev->iov[1].iov_base, &amp;index, &amp;tuple_arity)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > ev->iov[1].iov_len) goto error;
    if (tuple_arity != 2) goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_ulong(ev->iov[1].iov_base, &amp;index,
        &amp;main_term_binary_size)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > ev->iov[1].iov_len) goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_ulong(ev->iov[1].iov_base, &amp;index,
        &amp;byref_binaries_sizes_list_binary_size)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > ev->iov[1].iov_len) goto error;
    </xsl:if>

    <!-- Locate the start of the main term binary -->
    if (index &lt; ev->iov[1].iov_len) {
        <!-- The two first binaries have been packed -->
        main_term_binary_vector_index = 1;
    } else if (ev->vsize >= 3) {
        main_term_binary_vector_index = 2;
        index = 0;
    <xsl:if test='$runtime-checks'>
    } else {
        goto error;
    </xsl:if>
    }
    main_term_binary_buffer_size = main_term_binary_size + index;

    main_term_binary_vector_start_index = index;

    <xsl:if test='$runtime-checks'>
    if (main_term_binary_buffer_size
            > ev->iov[main_term_binary_vector_index].iov_len) goto error;
    </xsl:if>

    <!-- Decode the start of the main term binary -->
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_version(ev->iov[main_term_binary_vector_index].iov_base,
        &amp;index, &amp;version)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > ev->iov[main_term_binary_vector_index].iov_len)
        goto error;
    </xsl:if>
        
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_tuple_header(ev->iov[main_term_binary_vector_index].iov_base,
        &amp;index, &amp;tuple_arity)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > ev->iov[main_term_binary_vector_index].iov_len)
        goto error;
    if (tuple_arity &lt; 1) goto error;
    </xsl:if>

    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_ulong(ev->iov[main_term_binary_vector_index].iov_base, &amp;index,
        &amp;command_id)
    <xsl:if test='$runtime-checks'>) goto error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (index > ev->iov[main_term_binary_vector_index].iov_len)
        goto error;
    </xsl:if>

    <!-- The number of bytes left to be read in the main term binary -->
    main_term_binary_size -= (index - main_term_binary_vector_start_index);

    <xsl:if test='$runtime-checks'>
    if (main_term_binary_size &lt;= 0) goto error;
    </xsl:if>

    switch (command_id) {
<xsl:for-each select='dryverl:def-erlang-to-c-binding[dryverl:encode-input/dryverl:encode-input-byref-binary-list]'>
    case <xsl:value-of select='$constant-prefix'/>_COMMAND_<xsl:value-of select='@name'/>:
        <xsl:if test='$runtime-checks'>
        if (tuple_arity != (1 <xsl:if test='dryverl:encode-output'>+2</xsl:if>
                <xsl:if test='dryverl:encode-input/dryverl:encode-input-main-term'>+1</xsl:if>))
            goto error;
        </xsl:if>
        <xsl:value-of select='$prefix'/>_decode_input_<xsl:value-of select='@name'/>(
            driver_data, index, main_term_binary_buffer_size,
            ev, main_term_binary_vector_index, byref_binaries_sizes_list_binary_size,
            NULL<xsl:if test='dryverl:encode-output'>, NULL, 0, NULL</xsl:if>);
        return;
        break;
</xsl:for-each>
    <xsl:if test='$runtime-checks'>
    default:
        goto error;
        break;	
    </xsl:if>
    }

    return;

<xsl:if test='$runtime-checks'>
error:
    driver_failure_atom(driver_data->port, "badarg");
    return;
</xsl:if>

}
</xsl:if>

static void
<xsl:value-of select='$prefix'/>_ready_async(ErlDrvData drv_data, ErlDrvThreadData thread_data)
{
}

<!-- These constructs allows to specify a macro name in the driver's
     "name" attribute, instead of the real name. This allows to specify the
     real name only at compile time, e.g. by passing an option such as
     -DNAME=real_name to the compiler. The trick is to have two levels of macros
     to avoid the prescan of macros performed by the C preprocessor. -->
#define <xsl:value-of select='$constant-prefix'/>_DRIVER_NAME <xsl:value-of select='dryverl:gen-driver/@name'/>
#define <xsl:value-of select='$constant-prefix'/>_DRIVER_NAME_XSTR(NAME) \
    <xsl:value-of select='$constant-prefix'/>_DRIVER_NAME_STR(NAME)
#define <xsl:value-of select='$constant-prefix'/>_DRIVER_NAME_STR(NAME) #NAME
#define DRIVER_XINIT(DRIVER_NAME) DRIVER_INIT(DRIVER_NAME)

static ErlDrvEntry <xsl:value-of select='$prefix'/>_driver_entry = {
    <xsl:value-of select='$prefix'/>_init, /* init */
    <xsl:value-of select='$prefix'/>_start, /* start */
    <xsl:value-of select='$prefix'/>_stop, /* stop */
    NULL, /* output */
    NULL, /* ready_input */
    NULL, /* ready_output */
    <xsl:value-of select='$constant-prefix'/>_DRIVER_NAME_XSTR(<xsl:value-of select='$constant-prefix'/>_DRIVER_NAME), /* driver_name */
    NULL, /* finish */
    NULL, /* handle */
    NULL, /* control */
    NULL, /* timeout */
<xsl:choose>
<xsl:when test='dryverl:def-erlang-to-c-binding[dryverl:encode-input/dryverl:encode-input-byref-binary-list]'>
<xsl:text>    </xsl:text><xsl:value-of select='$prefix'/>_outputv, /* outputv */
</xsl:when>
<xsl:otherwise>
    NULL, /* outputv */
</xsl:otherwise>
</xsl:choose>
    <xsl:value-of select='$prefix'/>_ready_async, /* ready_async */
    NULL, /* flush */
    <xsl:value-of select='$prefix'/>_call, /* call */
    NULL /* event */
};

DRIVER_XINIT(<xsl:value-of select='$constant-prefix'/>_DRIVER_NAME)
{
    return &amp;<xsl:value-of select='$prefix'/>_driver_entry;
}


<xsl:for-each select='dryverl:def-value-map'>
<xsl:variable name='value-map-t'><xsl:value-of select='$prefix'/>_value_map_<xsl:value-of select='@name'/>_t</xsl:variable>
<xsl:variable name='value-map-entry-t'><xsl:value-of select='$prefix'/>_value_map_entry_<xsl:value-of select='@name'/>_t</xsl:variable>

<xsl:if test='dryverl:def-id-hash-map'>
<!-- This is Pr. D.J. Bernstein's hash function, in the public domain -->
unsigned int <xsl:value-of select='$prefix'/>_value_map_get_id_hash_for_<xsl:value-of select='@name'/>(unsigned long id) {
    unsigned int i;
    unsigned int h = 5381;
    unsigned char *buf = (unsigned char *) &amp;id;
    for (i = 0; i &lt; sizeof(id); i++) {
        h = (h + (h &lt;&lt; 5)) ^ buf[i];
    }
    return h % <xsl:value-of select='dryverl:def-id-hash-map/@table-size'/>;
}

void <xsl:value-of select='$prefix'/>_value_map_remove_id_hash_entry_from_<xsl:value-of select='@name'/>(<xsl:value-of select='$value-map-t'/> *value_map, unsigned int hash, <xsl:value-of select='$value-map-entry-t'/> *entry) {
    if (entry->previous_same_id_hash) {
        entry->previous_same_id_hash->next_same_id_hash =
            entry->next_same_id_hash;
    } else {
        value_map->id_hash_table[hash] = entry->next_same_id_hash;
    }
    if (entry->next_same_id_hash) {
        entry->next_same_id_hash->previous_same_id_hash =
            entry->previous_same_id_hash;
    }
    entry->previous_same_id_hash = NULL;
    entry->next_same_id_hash = NULL;
}
</xsl:if>

<xsl:if test='dryverl:def-value-hash-map'>
<!-- This is Pr. D.J. Bernstein's hash function, in the public domain -->
unsigned int <xsl:value-of select='$prefix'/>_value_map_get_value_hash_for_<xsl:value-of select='@name'/>(<xsl:value-of select='@c-type'/> *value) {
    unsigned int i;
    unsigned int h = 5381;
    unsigned char *buf = (unsigned char *) value;
    for (i = 0; i &lt; sizeof(<xsl:value-of select='@c-type'/>); i++) {
        h = (h + (h &lt;&lt; 5)) ^ buf[i];
    }
    return h % <xsl:value-of select='dryverl:def-value-hash-map/@table-size'/>;
}

void <xsl:value-of select='$prefix'/>_value_map_remove_value_hash_entry_from_<xsl:value-of select='@name'/>(<xsl:value-of select='$value-map-t'/> *value_map, unsigned int hash, <xsl:value-of select='$value-map-entry-t'/> *entry) {
    if (entry->previous_same_value_hash) {
        entry->previous_same_value_hash->next_same_value_hash =
            entry->next_same_value_hash;
    } else {
        value_map->value_hash_table[hash] = entry->next_same_value_hash;
    }
    if (entry->next_same_value_hash) {
        entry->next_same_value_hash->previous_same_value_hash =
            entry->previous_same_value_hash;
    }
    entry->previous_same_value_hash = NULL;
    entry->next_same_value_hash = NULL;
}
</xsl:if>

<!-- Remove the entry from the list of "used" entries, and put it into the list
     of "unused" entries -->
void <xsl:value-of select='$prefix'/>_value_map_remove_entry_<xsl:value-of select='@name'/>(<xsl:value-of select='$value-map-t'/> *value_map, <xsl:value-of select='$value-map-entry-t'/> *entry) {
    if (entry->previous) {
        entry->previous->next = entry->next;
    } else {
        value_map->first_used_entry = entry->next;
    }
    if (entry->next) {
        entry->next->previous = entry->previous;
    }
    <!-- If the entry was driver_alloc()ed, driver_free() it.
         Otherwise, put it back into the "unused" list. -->
    if (entry->flags &amp; <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAG_ALLOCED) {
        driver_free(entry);
    } else {
        entry->previous = NULL;
        entry->next = value_map->first_unused_entry;
        value_map->first_unused_entry = entry;
        if (entry->next) {
            entry->next->previous = entry;
        }
    }
}

void <xsl:value-of select='$prefix'/>_value_map_change_entry_flags_<xsl:value-of select='@name'/>(<xsl:value-of select='$value-map-entry-t'/> *entry, unsigned char flags_mask_or, unsigned flags_mask_xor) {
<xsl:if test='$runtime-checks'>
    if (!entry) return; <!-- FIXME: signal error somehow -->
</xsl:if>
    entry->flags |= (flags_mask_or &amp; <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAGS_PUBLIC_MASK);
    entry->flags ^= (flags_mask_xor &amp; <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAGS_PUBLIC_MASK);
}

int <xsl:value-of select='$prefix'/>_value_map_add_entry_<xsl:value-of select='@name'/>(<xsl:value-of select='$value-map-t'/> *value_map, unsigned long *id, <xsl:value-of select='@c-type'/> *value, <xsl:if test='dryverl:def-value-hash-map'>unsigned int value_hash, </xsl:if><xsl:value-of select='$value-map-entry-t'/> **entry_if_created) {
    <xsl:value-of select='$value-map-entry-t'/> *new_entry;
    unsigned long last_new_id;
    unsigned long new_id;
    <xsl:value-of select='$value-map-entry-t'/> *used_entry;
<xsl:if test='dryverl:def-id-hash-map'>
    unsigned int id_hash;
</xsl:if>

    <!-- This function must be called only when the value is not already mapped
         in an entry.  Since this function is not public, and is called only by
         ..._value_map_find_entry_..., there is little chance that this
         happens. -->

    <xsl:if test='$runtime-checks'>
    if ((!value_map) || (!value)) return -EINVAL;
    </xsl:if>

    <!-- Find an unused id. Collisions should rarely occur (only when the
         value_map->next_id wraps around, and some old entries are kept for a
         long time in the map). -->
    new_id = value_map->next_id++;
    last_new_id = new_id - 1;
    while(1) {
        <!-- Check that the id is not allocated to a currently used entry. -->
        used_entry = value_map->first_used_entry;
        while(used_entry &amp;&amp; (used_entry->id == new_id)) {
            used_entry = used_entry->next;
        }
        if (!used_entry) {
            break;
        }
        new_id = value_map->next_id;
	<!-- As a safety measure, make sure that the sign bit is never set to
	     1, so that even if programs access this value as a signed long,
	     the value is still always >= 0 -->
        <!-- FIXME: make that modulo configurable in the XML document -->
	value_map->next_id = (new_id + 1) % (1UL + LONG_MAX);
        if (new_id == last_new_id) {
            return -ENOMEM;
        }
    }

    <!-- Try to use one of the pre-allocated entries. Otherwise, driver_alloc()
         a new one, and set the ALLOCED flag. -->
    new_entry = value_map->first_unused_entry;
    if (new_entry) {
        <!-- Remove from "unused" list -->
        value_map->first_unused_entry = new_entry->next;
        if (new_entry->next) {
            new_entry->next->previous = NULL;
        }
        <!-- Clear all flags, especially the ALLOCED flag. -->
        new_entry->flags = 0;
    } else {
        new_entry = (<xsl:value-of select='$value-map-entry-t'/>*)driver_alloc(
            sizeof(<xsl:value-of select='$value-map-entry-t'/>));
        if (!new_entry) {
            return -ENOMEM;
        }
        new_entry->flags =
            <xsl:value-of select='$constant-prefix'/>_VALUE_MAP_ENTRY_FLAG_ALLOCED;
    }

    new_entry->id = new_id;
    if (id) {
        *id = new_id;
    }
    new_entry->value = *value;
    <!-- Put into "used" list -->
    new_entry->next = value_map->first_used_entry;
    value_map->first_used_entry = new_entry;
    if (new_entry->next) {
        new_entry->next->previous = new_entry;
    }
<xsl:if test='dryverl:def-id-hash-map'>
    <!-- Put into id hash table -->
    id_hash = <xsl:value-of select='$prefix'/>_value_map_get_id_hash_for_<xsl:value-of select='@name'/>(new_id);
    new_entry->next_same_id_hash = value_map->id_hash_table[id_hash];
    if (new_entry->next_same_id_hash) {
        new_entry->next_same_id_hash->previous_same_id_hash = new_entry;
    }
    new_entry->previous_same_id_hash = NULL;
    value_map->id_hash_table[id_hash] = new_entry;
</xsl:if>
<xsl:if test='dryverl:def-value-hash-map'>
    <!-- Put into value hash table -->
    new_entry->next_same_value_hash = value_map->value_hash_table[value_hash];
    if (new_entry->next_same_value_hash) {
        new_entry->next_same_value_hash->previous_same_value_hash = new_entry;
    }
    new_entry->previous_same_value_hash = NULL;
    value_map->value_hash_table[value_hash] = new_entry;
</xsl:if>
    if (entry_if_created) {
        *entry_if_created = new_entry;
    }
    return 0;
}

<xsl:if test='dryverl:def-value-destroy-code'>
<!-- Put that code in a separate function to prevent the code to interfere with
     the generated code (identifiers conflicts, etc.). -->
void <xsl:value-of select='$prefix'/>_value_map_destroy_value_<xsl:value-of select='@name'/>(<xsl:value-of select='@c-type'/> *value) {
    <xsl:apply-templates select='dryverl:def-value-destroy-code'/>
}
</xsl:if>

int <xsl:value-of select='$prefix'/>_value_map_find_entry_<xsl:value-of select='@name'/>(<xsl:value-of select='$value-map-t'/> *value_map, <xsl:value-of select='$prefix'/>_value_map_entry_op_t entry_op, unsigned long *id, <xsl:value-of select='@c-type'/> *value, unsigned char *flags_before_modif, unsigned char flags_mask_or, unsigned char flags_mask_xor) {
<xsl:if test='dryverl:def-id-hash-map or dryverl:def-value-hash-map'>
    unsigned int hash;
</xsl:if>
    <xsl:value-of select='$value-map-entry-t'/> *entry;
    int temp_id = 0;

    <xsl:if test='$runtime-checks'>
    if (!value_map) return -EINVAL;
    </xsl:if>

    switch(entry_op) {
    case FIND_ID:
    case REMOVE_ID:
    case DESTROY_ID:
        <!-- Find the entry according the given id. -->
        <xsl:if test='$runtime-checks'>
        if (!id) return -EINVAL;
        </xsl:if>
        temp_id = *id;
<xsl:choose>
<xsl:when test='dryverl:def-id-hash-map'>
        hash = <xsl:value-of select='$prefix'/>_value_map_get_id_hash_for_<xsl:value-of select='@name'/>(temp_id);
        entry = value_map->id_hash_table[hash];
        while (entry &amp;&amp; (entry->id != temp_id)) {
            entry = entry->next_same_id_hash;
        }
</xsl:when>
<xsl:otherwise>
        entry = value_map->first_used_entry;
        while (entry &amp;&amp; (entry->id != temp_id)) {
            entry = entry->next;
        }
</xsl:otherwise>
</xsl:choose>
        if (!entry) {
            return 1; <!-- Non-error code for "not found". -->
        }
        if (value) {
            *value = entry->value;
        }
        break;
    case FIND_VALUE:
    case FIND_VALUE_AND_ADD_IF_NOT_FOUND:
    case REMOVE_VALUE:
    case DESTROY_VALUE:
        <!-- Find the entry according the given value. -->
        <xsl:if test='$runtime-checks'>
        if (!value) return -EINVAL;
        </xsl:if>
<xsl:choose>
<xsl:when test='dryverl:def-value-hash-map'>
        hash = <xsl:value-of select='$prefix'/>_value_map_get_value_hash_for_<xsl:value-of select='@name'/>(value);
        entry = value_map->value_hash_table[hash];
        while (entry &amp;&amp; memcmp(&amp;entry->value, value,
                                       sizeof(<xsl:value-of select='@c-type'/>))) {
            entry = entry->next_same_value_hash;
        }
</xsl:when>
<xsl:otherwise>
        entry = value_map->first_used_entry;
        while (entry &amp;&amp; memcmp(&amp;entry->value, value,
                                       sizeof(<xsl:value-of select='@c-type'/>))) {
            entry = entry->next;
        }
</xsl:otherwise>
</xsl:choose>
        if (!entry) {
            if (entry_op == FIND_VALUE_AND_ADD_IF_NOT_FOUND) {
                <xsl:if test='$runtime-checks'>
                int return_value =
                </xsl:if>
                <xsl:value-of select='$prefix'/>_value_map_add_entry_<xsl:value-of select='@name'/>(value_map, id, value, <xsl:if test='dryverl:def-value-hash-map'>hash, </xsl:if>&amp;entry);
                <xsl:if test='$runtime-checks'>
                if (return_value) return return_value;
                </xsl:if>
            } else {
                return 1; <!-- Non-error code for "not found". -->
            }
        }
        if (id) {
            *id = entry->id;
        }
        break;
    <xsl:if test='$runtime-checks'>
    default:
        return -EINVAL;
        break;
    </xsl:if>
    }

    if (flags_before_modif) {
        *flags_before_modif = entry->flags;
    }

<xsl:if test='dryverl:def-value-destroy-code'>
    switch(entry_op) {
    case DESTROY_ID:
    case DESTROY_VALUE:
        <xsl:value-of select='$prefix'/>_value_map_destroy_value_<xsl:value-of select='@name'/>(&amp;entry->value);
        break;
    default:
        break;
    }
</xsl:if>

    switch(entry_op) {
    case REMOVE_ID:
    case DESTROY_ID:
<xsl:if test='dryverl:def-id-hash-map'>
        <xsl:value-of select='$prefix'/>_value_map_remove_id_hash_entry_from_<xsl:value-of select='@name'/>(value_map, hash, entry);
</xsl:if>
<xsl:if test='dryverl:def-value-hash-map'>
        hash = <xsl:value-of select='$prefix'/>_value_map_get_value_hash_for_<xsl:value-of select='@name'/>(value);
        <xsl:value-of select='$prefix'/>_value_map_remove_value_hash_entry_from_<xsl:value-of select='@name'/>(value_map, hash, entry);
</xsl:if>
        <xsl:value-of select='$prefix'/>_value_map_remove_entry_<xsl:value-of select='@name'/>(value_map, entry);
        break;
    case REMOVE_VALUE:
    case DESTROY_VALUE:
<xsl:if test='dryverl:def-value-hash-map'>
        <xsl:value-of select='$prefix'/>_value_map_remove_value_hash_entry_from_<xsl:value-of select='@name'/>(value_map, hash, entry);
</xsl:if>
<xsl:if test='dryverl:def-id-hash-map'>
        hash = <xsl:value-of select='$prefix'/>_value_map_get_id_hash_for_<xsl:value-of select='@name'/>(entry->id);
        <xsl:value-of select='$prefix'/>_value_map_remove_id_hash_entry_from_<xsl:value-of select='@name'/>(value_map, hash, entry);
</xsl:if>
        <xsl:value-of select='$prefix'/>_value_map_remove_entry_<xsl:value-of select='@name'/>(value_map, entry);
        break;
    default:
        <xsl:value-of select='$prefix'/>_value_map_change_entry_flags_<xsl:value-of select='@name'/>(entry, flags_mask_or, flags_mask_xor);
        break;
    }
    return 0;
}
</xsl:for-each>

</xsl:template>

<!-- Templates for common elements that can be used in C code -->

<xsl:template match='dryverl:c-local-variable'>
<xsl:if test='not(ancestor::*/dryverl:def-c-local-variable/@name = @name)'>
<xsl:message terminate="yes">
Access to an undefined local variable "<xsl:value-of select='@name'/>".
</xsl:message>
</xsl:if>
(local_variable_<xsl:value-of select='@name'/>)
</xsl:template>

<xsl:template match='dryverl:c-call-variable'>
<xsl:if test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:def-c-call-variable/@name = @name)'>
<xsl:message terminate="yes">
Access to an undefined call variable "<xsl:value-of select='@name'/>".
</xsl:message>
</xsl:if>
(call_variables-><xsl:value-of select='@name'/>)
</xsl:template>

<xsl:template match='dryverl:failure-atom'>
driver_failure_atom(call_state->driver_data->port, (<xsl:apply-templates/>));
return;
</xsl:template>

<xsl:template match='dryverl:failure-posix'>
driver_failure_posix(call_state->driver_data->port, (<xsl:apply-templates/>));
return;
</xsl:template>

<xsl:template match='dryverl:failure'>
driver_failure(call_state->driver_data->port, (<xsl:apply-templates/>));
return;
</xsl:template>

<!-- Templates for elements that can be used in <decode-input-main-term/>
     elements -->

<!-- Templates for decoding binaries passed by reference -->

<xsl:template match='dryverl:byref-binaries-count'>
<xsl:if test='not(ancestor::dryverl:assign-c-call-variables)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;assign-c-call-variables/&gt;.
</xsl:message>
</xsl:if>
<xsl:if test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:encode-input/dryverl:encode-input-byref-binary-list)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> in a &lt;def-erlang-to-c-binding/> containing no &lt;encode-input-byref-binary-list/>.
</xsl:message>
</xsl:if>
(byref_binaries_count)
</xsl:template>

<xsl:template match='dryverl:decode-input-byref-binary'>
<xsl:if test='not(ancestor::dryverl:assign-c-call-variables)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;assign-c-call-variables/&gt;.
</xsl:message>
</xsl:if>
<xsl:if test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:encode-input/dryverl:encode-input-byref-binary-list)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> in a &lt;def-erlang-to-c-binding/> containing no &lt;encode-input-byref-binary-list/>.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_byref_binaries_count-- &lt;= 0) goto decoding_error;
    </xsl:if>
    <!-- First, get the size of the buffer to "decode", from the list.
         This is done differently if the list is encoded as a real list, or as
	 a string. -->
    if (byref_binaries_sizes_list_is_a_string) {
        temp_byref_binary_size =
            ((unsigned char*)byref_binaries_sizes_list_buffer)
                [byref_binaries_sizes_list_buffer_index++];
    } else {
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_decode_ulong(byref_binaries_sizes_list_buffer,
            &amp;byref_binaries_sizes_list_buffer_index,
            &amp;temp_byref_binary_size)
        <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
    }
    <xsl:if test='$runtime-checks'>
    if (byref_binaries_sizes_list_buffer_index
            > byref_binaries_sizes_list_buffer_size) goto decoding_error;
    if (temp_byref_binary_size &lt; 0) goto decoding_error;
    if ((byref_binary_buffer_index + temp_byref_binary_size)
            > ev->iov[byref_binary_vector_index].iov_len) goto decoding_error;
    </xsl:if>
    <!-- Copy the decoded data into the user-provided variables -->
    *<!--(char**)-->(<xsl:apply-templates select='dryverl:decode-input-byref-binary-buf-ref-into'/>) =
        ev->iov[byref_binary_vector_index].iov_base;
    *<!--(ErlDrvBinary**)-->(<xsl:apply-templates select='dryverl:decode-input-byref-binary-ref-into'/>) =
        ev->binv[byref_binary_vector_index];
    *<!--(unsigned long*)-->(<xsl:apply-templates select='dryverl:decode-input-byref-binary-size-into'/>) =
        temp_byref_binary_size;
    *<!--(unsigned long*)-->(<xsl:apply-templates select='dryverl:decode-input-byref-binary-offset-into'/>) =
        byref_binary_buffer_index;
    <!-- Find the pointer to the next byref binary -->
    <xsl:if test='$runtime-checks'>
    if (allowed_byref_binaries_count > 0) {
    </xsl:if>
    if ((byref_binary_buffer_index + temp_byref_binary_size)
            == ev->iov[byref_binary_vector_index].iov_len) {
        byref_binary_vector_index++;
        byref_binary_buffer_index = 0;
        <xsl:if test='$runtime-checks'>
        if (byref_binary_vector_index >= ev->vsize) goto decoding_error;
        </xsl:if>
    } else {
        byref_binary_buffer_index += temp_byref_binary_size;
    }
    <xsl:if test='$runtime-checks'>
    }
    </xsl:if>
</xsl:template>

<!-- Templates for decoding terms passed by copy -->

<!-- NOTE: Since there are no buffer boundary checks in ei_decode_* functions,
     we can only check *after* decoding if there has been an access out of the
     buffer's boundaries, i.e. if something really bad has just happened. -->

<xsl:template name='decode-input-generic-arity1'>
<xsl:param name='c-type'/>
<xsl:param name='ei-decode-suffix'/>
<xsl:if test='not(ancestor::dryverl:assign-c-call-variables)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;assign-c-call-variables/&gt;.
</xsl:message>
</xsl:if>
<xsl:if test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:encode-input/dryverl:encode-input-main-term)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> in a &lt;def-erlang-to-c-binding/> containing no &lt;encode-input-main-term/>.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count-- &lt;= 0) goto decoding_error;
    </xsl:if>
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_<xsl:value-of select='$ei-decode-suffix'/>(main_term_buffer,
        &amp;main_term_buffer_index, <!--(<xsl:value-of select='$c-type'/>*)-->
        (<xsl:apply-templates/>))
    <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto decoding_error;
    </xsl:if>
</xsl:template>

<xsl:template name='decode-input-generic-arity2'>
<xsl:param name='c-type1'/>
<xsl:param name='child-node1'/>
<xsl:param name='c-type2'/>
<xsl:param name='child-node2'/>
<xsl:param name='ei-decode-suffix'/>
<xsl:if test='not(ancestor::dryverl:assign-c-call-variables)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;assign-c-call-variables/&gt;.
</xsl:message>
</xsl:if>
<xsl:if test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:encode-input/dryverl:encode-input-main-term)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> in a &lt;def-erlang-to-c-binding/> containing no &lt;encode-input-main-term/>.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count-- &lt;= 0) goto decoding_error;
    </xsl:if>
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_decode_<xsl:value-of select='$ei-decode-suffix'/>(main_term_buffer,
        &amp;main_term_buffer_index,
        <!--(<xsl:value-of select='$c-type1'/>*)-->
            (<xsl:apply-templates select='$child-node1'/>),
        <!--(<xsl:value-of select='$c-type2'/>*)-->
            (<xsl:apply-templates select='$child-node2'/>))
    <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
    <xsl:if test='$runtime-checks'>
    if (main_term_buffer_index > main_term_buffer_size) goto decoding_error;
    </xsl:if>
</xsl:template>

<xsl:template name='decode-input-generic-compound'>
<xsl:param name='size-child-node'/>
<xsl:param name='contents-child-node'/>
<xsl:param name='ei-decode-suffix'/>
<xsl:param name='add-to-terms-count' select='0'/>
<xsl:if test='not(ancestor::dryverl:assign-c-call-variables)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;assign-c-call-variables/&gt;.
</xsl:message>
</xsl:if>
<xsl:if test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:encode-input/dryverl:encode-input-main-term)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> in a &lt;def-erlang-to-c-binding/> containing no &lt;encode-input-main-term/>.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count-- &lt;= 0) goto decoding_error;
    </xsl:if>
    {
        int allowed_terms_count = 0;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_decode_<xsl:value-of select='$ei-decode-suffix'/>(main_term_buffer,
            &amp;main_term_buffer_index, &amp;allowed_terms_count)
        <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
        <xsl:if test='$runtime-checks'>
        if (main_term_buffer_index > main_term_buffer_size) goto decoding_error;
        </xsl:if>
        *(<xsl:apply-templates select='$size-child-node'/>) = allowed_terms_count;
	<xsl:if test='$runtime-checks and ($add-to-terms-count > 0)'>
	allowed_terms_count += <xsl:value-of select='$add-to-terms-count'/>;
        </xsl:if>
        {<xsl:apply-templates select='$contents-child-node'/>}
        <xsl:if test='$runtime-checks'>
        if (allowed_terms_count != 0) goto decoding_error;
        </xsl:if>
    }
</xsl:template>

<xsl:template match='dryverl:decode-input-long-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>long</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>long</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-ulong-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>unsigned long</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>ulong</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-double-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>double</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>double</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-boolean-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>int</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>boolean</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-char-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>char</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>char</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-string-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>char</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>string</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-atom-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>char</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>atom</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-binary'>
<xsl:call-template name='decode-input-generic-arity2'>
<xsl:with-param name='c-type1'>void</xsl:with-param>
<xsl:with-param name='child-node1' select='dryverl:decode-input-binary-data-into'/>
<xsl:with-param name='c-type2'>long</xsl:with-param>
<xsl:with-param name='child-node2' select='dryverl:decode-input-binary-size-into'/>
<xsl:with-param name='ei-decode-suffix'>binary</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-fun-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>erlang_fun</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>fun</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-pid-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>erlang_pid</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>pid</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-port-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>erlang_port</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>port</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-ref-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>erlang_ref</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>ref</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-trace-into'>
<xsl:call-template name='decode-input-generic-arity1'>
<xsl:with-param name='c-type'>erlang_trace</xsl:with-param>
<xsl:with-param name='ei-decode-suffix'>trace</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-tuple'>
<xsl:call-template name='decode-input-generic-compound'>
<xsl:with-param name='size-child-node' select='dryverl:decode-input-tuple-size-into'/>
<xsl:with-param name='contents-child-node' select='dryverl:decode-input-tuple-contents'/>
<xsl:with-param name='ei-decode-suffix'>tuple_header</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-list'>
<xsl:call-template name='decode-input-generic-compound'>
<xsl:with-param name='size-child-node' select='dryverl:decode-input-list-size-into'/>
<xsl:with-param name='contents-child-node' select='dryverl:decode-input-list-contents'/>
<xsl:with-param name='ei-decode-suffix'>list_header</xsl:with-param>
<xsl:with-param name='add-to-terms-count' select='1'/> <!-- for the tail -->
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:decode-input-next-term-info'>
<xsl:if test='not(ancestor::dryverl:assign-c-call-variables)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;assign-c-call-variables/&gt;.
</xsl:message>
</xsl:if>
<xsl:if test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:encode-input/dryverl:encode-input-main-term)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> in a &lt;def-erlang-to-c-binding/> containing no &lt;encode-input-main-term/>.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count &lt;= 0) { goto decoding_error; }
    </xsl:if>
    <xsl:if test='$runtime-checks'>if (</xsl:if>
    ei_get_type(main_term_buffer, &amp;main_term_buffer_index,
        <!--(int*)-->(<xsl:apply-templates select='dryverl:decode-input-next-term-type-into'/>),
        <!--(int*)-->(<xsl:apply-templates select='dryverl:decode-input-next-term-size-into'/>))
    <xsl:if test='$runtime-checks'>) goto decoding_error</xsl:if>;
</xsl:template>


<!-- Templates for elements that can be used in <encode-output-main-term/>
     elements -->

<!-- Template for encoding binaries passed by reference -->

<xsl:template match='dryverl:encode-output-byref-binary'>
<xsl:if test='not(ancestor::dryverl:encode-output-main-term-and-byref-binaries)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;encode-output-main-term-and-byref-binaries/&gt;.
</xsl:message>
</xsl:if>
    if (! driver_term) {
        <!-- This is the first encoded byref binary -->
        driver_term_size = 128;
        driver_term = (ErlDrvTermData*) driver_alloc(
            sizeof(ErlDrvTermData) * driver_term_size);
        if (! main_term_binary) {
            main_term_binary = driver_alloc_binary(main_term_buffer_size);
            <xsl:if test='$runtime-checks'>
            if (! main_term_binary) goto memory_error;
            </xsl:if>
            memcpy(main_term_binary->orig_bytes, main_term_buffer,
                main_term_buffer_index);
            if (main_term_buffer_was_alloced) {
                driver_free(main_term_buffer);
            }
            main_term_buffer = main_term_binary->orig_bytes;
        }
    } else if ((driver_term_index + 9) >= driver_term_size) {
        <!-- We add 9, for the binary is encoded in 4 indexes, and we need to
             keep 5 indexes at the end to encode the list and tuple headers -->
        driver_term_size += 128;
        driver_term = (ErlDrvTermData*) driver_realloc(driver_term,
            sizeof(ErlDrvTermData) * driver_term_size);
    }
    <xsl:if test='$runtime-checks'>
    if (! driver_term) goto memory_error;
    </xsl:if>
    driver_term[driver_term_index++] = ERL_DRV_BINARY;
    <!-- FIXME: should check the user-provided values -->
    driver_term[driver_term_index++] = (ErlDrvTermData) <!--(ErlDrvBinary*)-->
        (<xsl:apply-templates select='dryverl:encode-output-byref-binary-ref'/>);
    driver_term[driver_term_index++] = (ErlDrvTermData) <!--(unsigned int)-->
        (<xsl:apply-templates select='dryverl:encode-output-byref-binary-size'/>);
    driver_term[driver_term_index++] = (ErlDrvTermData) <!--(unsigned int)-->
        (<xsl:apply-templates select='dryverl:encode-output-byref-binary-offset'/>);
    byref_binaries_count++;
</xsl:template>

<!-- Templates for encoding terms passed by copy -->

<!-- Factorizes the templates for commands that encode no values -->
<xsl:template name='encode-output-generic-arity0'>
<xsl:param name='ei-encode-suffix'/>
<xsl:if test='not(ancestor::dryverl:encode-output-main-term-and-byref-binaries)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;encode-output-main-term-and-byref-binaries/&gt;.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count-- &lt;= 0) { goto encoding_error; }
    </xsl:if>
    {
        temp_index = main_term_buffer_index;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            NULL, &amp;temp_index)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
        if (temp_index > main_term_buffer_size) {
            old_buffer_size = main_term_buffer_size;
            main_term_buffer_size = temp_index + 100;
            if (main_term_binary) {
                main_term_binary = driver_realloc_binary(main_term_binary,
                    main_term_buffer_size);
                <xsl:if test='$runtime-checks'>
                if (! main_term_binary) goto memory_error;
                </xsl:if>
                main_term_buffer = main_term_binary->orig_bytes;
            } else if (main_term_buffer_was_alloced) {
                main_term_buffer = (char*) driver_realloc(main_term_buffer,
                    main_term_buffer_size);
            } else {
                temp_buffer = (char*) driver_alloc(
                    main_term_buffer_size);
                memcpy(temp_buffer, main_term_buffer,
                    old_buffer_size);
                main_term_buffer = temp_buffer;
                main_term_buffer_was_alloced = 1;
            }
            <xsl:if test='$runtime-checks'>
            if (! main_term_buffer) goto memory_error;
            </xsl:if>
        }
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            main_term_buffer, &amp;main_term_buffer_index)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
    }
</xsl:template>

<!-- Factorizes the templates for commands that encode single values -->
<xsl:template name='encode-output-generic-arity1'>
<xsl:param name='c-type'/>
<xsl:param name='ei-encode-suffix'/>
<xsl:if test='not(ancestor::dryverl:encode-output-main-term-and-byref-binaries)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;encode-output-main-term-and-byref-binaries/&gt;.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count-- &lt;= 0) { goto encoding_error; }
    </xsl:if>
    {
        <xsl:value-of select='$c-type'/> temp_value =
            <!--(<xsl:value-of select='$c-type'/>)-->(<xsl:apply-templates/>);
        temp_index = main_term_buffer_index;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            NULL, &amp;temp_index, temp_value)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
        if (temp_index > main_term_buffer_size) {
            old_buffer_size = main_term_buffer_size;
            main_term_buffer_size = temp_index + 100;
            if (main_term_binary) {
                main_term_binary = driver_realloc_binary(main_term_binary,
                    main_term_buffer_size);
                <xsl:if test='$runtime-checks'>
                if (! main_term_binary) goto memory_error;
                </xsl:if>
                main_term_buffer = main_term_binary->orig_bytes;
            } else if (main_term_buffer_was_alloced) {
                main_term_buffer = (char*) driver_realloc(main_term_buffer,
                    main_term_buffer_size);
            } else {
                temp_buffer = (char*) driver_alloc(
                    main_term_buffer_size);
                memcpy(temp_buffer, main_term_buffer,
                    old_buffer_size);
                main_term_buffer = temp_buffer;
                main_term_buffer_was_alloced = 1;
            }
            <xsl:if test='$runtime-checks'>
            if (! main_term_buffer) goto memory_error;
            </xsl:if>
        }
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            main_term_buffer, &amp;main_term_buffer_index, temp_value)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
    }
</xsl:template>

<!-- Factorizes the templates for commands that encode two values -->
<xsl:template name='encode-output-generic-arity2'>
<xsl:param name='c-type1'/>
<xsl:param name='child-node1'/>
<xsl:param name='c-type2'/>
<xsl:param name='child-node2'/>
<xsl:param name='ei-encode-suffix'/>
<xsl:if test='not(ancestor::dryverl:encode-output-main-term-and-byref-binaries)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;encode-output-main-term-and-byref-binaries/&gt;.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count-- &lt;= 0) { goto encoding_error; }
    </xsl:if>
    {
        <xsl:value-of select='$c-type1'/> temp_value1 =
            <!--(<xsl:value-of select='$c-type1'/>)-->
            (<xsl:apply-templates select='$child-node1'/>);
        <xsl:value-of select='$c-type2'/> temp_value2 =
            <!--(<xsl:value-of select='$c-type2'/>)-->
            (<xsl:apply-templates select='$child-node2'/>);
        temp_index = main_term_buffer_index;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            NULL, &amp;temp_index, temp_value1, temp_value2)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
        if (temp_index > main_term_buffer_size) {
            old_buffer_size = main_term_buffer_size;
            main_term_buffer_size = temp_index + 100;
            if (main_term_binary) {
                main_term_binary = driver_realloc_binary(main_term_binary,
                    main_term_buffer_size);
                <xsl:if test='$runtime-checks'>
                if (! main_term_binary) goto memory_error;
                </xsl:if>
                main_term_buffer = main_term_binary->orig_bytes;
            } else if (main_term_buffer_was_alloced) {
                main_term_buffer = (char*) driver_realloc(main_term_buffer,
                    main_term_buffer_size);
            } else {
                temp_buffer = (char*) driver_alloc(
                    main_term_buffer_size);
                memcpy(temp_buffer, main_term_buffer,
                    old_buffer_size);
                main_term_buffer = temp_buffer;
                main_term_buffer_was_alloced = 1;
            }
            <xsl:if test='$runtime-checks'>
            if (! main_term_buffer) goto memory_error;
            </xsl:if>
        }
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            main_term_buffer, &amp;main_term_buffer_index, temp_value1, temp_value2)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
    }
</xsl:template>

<!-- Factorizes the templates for commands that encode tuples and lists -->
<xsl:template name='encode-output-generic-compound'>
<xsl:param name='size-child-node'/>
<xsl:param name='contents-child-node'/>
<xsl:param name='ei-encode-suffix'/>
<xsl:param name='add-to-terms-count' select='0'/>
<xsl:if test='not(ancestor::dryverl:encode-output-main-term-and-byref-binaries)'>
<xsl:message terminate="yes">
Illegal usage of a &lt;<xsl:value-of select='name(.)'/>/> outside of a &lt;encode-output-main-term-and-byref-binaries/&gt;.
</xsl:message>
</xsl:if>
    <xsl:if test='$runtime-checks'>
    if (allowed_terms_count-- &lt;= 0) { goto encoding_error; }
    </xsl:if>
    {
        int allowed_terms_count =
            <!--(int)-->(<xsl:apply-templates select='$size-child-node'/>);
        temp_index = main_term_buffer_index;
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            NULL, &amp;temp_index, allowed_terms_count)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
        if (temp_index > main_term_buffer_size) {
            old_buffer_size = main_term_buffer_size;
            main_term_buffer_size = temp_index + 100;
            if (main_term_binary) {
                main_term_binary = driver_realloc_binary(main_term_binary,
                    main_term_buffer_size);
                <xsl:if test='$runtime-checks'>
                if (! main_term_binary) goto memory_error;
                </xsl:if>
                main_term_buffer = main_term_binary->orig_bytes;
            } else if (main_term_buffer_was_alloced) {
                main_term_buffer = (char*) driver_realloc(main_term_buffer,
                    main_term_buffer_size);
            } else {
                temp_buffer = (char*) driver_alloc(
                    main_term_buffer_size);
                memcpy(temp_buffer, main_term_buffer,
                    old_buffer_size);
                main_term_buffer = temp_buffer;
                main_term_buffer_was_alloced = 1;
            }
            <xsl:if test='$runtime-checks'>
            if (! main_term_buffer) goto memory_error;
            </xsl:if>
        }
        <xsl:if test='$runtime-checks'>if (</xsl:if>
        ei_encode_<xsl:value-of select='$ei-encode-suffix'/>(
            main_term_buffer, &amp;main_term_buffer_index, allowed_terms_count)
        <xsl:if test='$runtime-checks'>) goto encoding_error</xsl:if>;
	<xsl:if test='$runtime-checks and ($add-to-terms-count > 0)'>
	allowed_terms_count += <xsl:value-of select='$add-to-terms-count'/>;
        </xsl:if>
        {<xsl:apply-templates select='$contents-child-node'/>}
        <xsl:if test='$runtime-checks'>
        if (allowed_terms_count != 0) { goto encoding_error; }
        </xsl:if>
    }
</xsl:template>

<xsl:template match='dryverl:encode-output-long'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>long</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>long</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-ulong'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>unsigned long</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>ulong</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-double'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>double</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>double</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-boolean'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>int</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>boolean</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-char'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>char</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>char</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-string'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>const char *</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>string</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-string-len'>
<xsl:call-template name='encode-output-generic-arity2'>
<xsl:with-param name='c-type1'>const char *</xsl:with-param>
<xsl:with-param name='child-node1' select='dryverl:encode-output-string-len-string'/>
<xsl:with-param name='c-type2'>int</xsl:with-param>
<xsl:with-param name='child-node2' select='dryverl:encode-output-string-len-size'/>
<xsl:with-param name='ei-encode-suffix'>string_len</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-atom'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>const char *</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>atom</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-atom-len'>
<xsl:call-template name='encode-output-generic-arity2'>
<xsl:with-param name='c-type1'>const char *</xsl:with-param>
<xsl:with-param name='child-node1' select='dryverl:encode-output-atom-len-string'/>
<xsl:with-param name='c-type2'>int</xsl:with-param>
<xsl:with-param name='child-node2' select='dryverl:encode-output-atom-len-size'/>
<xsl:with-param name='ei-encode-suffix'>atom_len</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-binary'>
<xsl:call-template name='encode-output-generic-arity2'>
<xsl:with-param name='c-type1'>const void *</xsl:with-param>
<xsl:with-param name='child-node1' select='dryverl:encode-output-binary-data'/>
<xsl:with-param name='c-type2'>long</xsl:with-param>
<xsl:with-param name='child-node2' select='dryverl:encode-output-binary-size'/>
<xsl:with-param name='ei-encode-suffix'>binary</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-pid'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>erlang_pid *</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>pid</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-fun'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>erlang_fun *</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>fun</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-port'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>erlang_port *</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>port</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-ref'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>erlang_ref *</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>ref</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-trace'>
<xsl:call-template name='encode-output-generic-arity1'>
<xsl:with-param name='c-type'>erlang_trace *</xsl:with-param>
<xsl:with-param name='ei-encode-suffix'>trace</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-tuple'>
<xsl:call-template name='encode-output-generic-compound'>
<xsl:with-param name='size-child-node' select='dryverl:encode-output-tuple-size'/>
<xsl:with-param name='contents-child-node' select='dryverl:encode-output-tuple-contents'/>
<xsl:with-param name='ei-encode-suffix'>tuple_header</xsl:with-param>
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-list'>
<xsl:call-template name='encode-output-generic-compound'>
<xsl:with-param name='size-child-node' select='dryverl:encode-output-list-size'/>
<xsl:with-param name='contents-child-node' select='dryverl:encode-output-list-contents'/>
<xsl:with-param name='ei-encode-suffix'>list_header</xsl:with-param>
<xsl:with-param name='add-to-terms-count' select='1'/> <!-- for the tail -->
</xsl:call-template>
</xsl:template>

<xsl:template match='dryverl:encode-output-empty-list'>
<xsl:call-template name='encode-output-generic-arity0'>
<xsl:with-param name='ei-encode-suffix'>empty_list</xsl:with-param>
</xsl:call-template>
</xsl:template>


<!-- Templates for elements usable in def-value-destroy-code elements. -->

<xsl:template match='dryverl:value-ref'>
(value)
</xsl:template>


<!-- Templates for operations manipulating value maps -->

<xsl:template match='dryverl:value-map-lock'>
<xsl:if test='not(ancestor::dryverl:def-bindings/dryverl:def-value-map/@name = @name)'>
<xsl:message terminate="yes">
Access to an undefined value map "<xsl:value-of select='@name'/>".
</xsl:message>
</xsl:if>
    pthread_mutex_lock(&amp;call_state->driver_data->value_map_<xsl:value-of select='@name'/>.mutex);
</xsl:template>

<xsl:template match='dryverl:value-map-unlock'>
<xsl:if test='not(ancestor::dryverl:def-bindings/dryverl:def-value-map/@name = @name)'>
<xsl:message terminate="yes">
Access to an undefined value map "<xsl:value-of select='@name'/>".
</xsl:message>
</xsl:if>
    pthread_mutex_unlock(&amp;call_state->driver_data->value_map_<xsl:value-of select='@name'/>.mutex);
</xsl:template>

<xsl:template match='dryverl:value-map-find-entry'>
<xsl:variable name='name' select='@name'/>
<xsl:if test='not(ancestor::dryverl:def-bindings/dryverl:def-value-map/@name = $name)'>
<xsl:message terminate="yes">
Access to an undefined value map "<xsl:value-of select='$name'/>".
</xsl:message>
</xsl:if>
    switch(<xsl:value-of select='$prefix'/>_value_map_find_entry_<xsl:value-of select='$name'/>(
        &amp;call_state->driver_data->value_map_<xsl:value-of select='$name'/>,
<xsl:choose>
<xsl:when test='dryverl:value-map-find-entry-id'>
            FIND_ID,
</xsl:when>
<xsl:when test='dryverl:value-map-find-entry-value'>
            FIND_VALUE,
</xsl:when>
<xsl:when test='dryverl:value-map-find-or-add-entry-value'>
            FIND_VALUE_AND_ADD_IF_NOT_FOUND,
</xsl:when>
<xsl:when test='dryverl:value-map-remove-entry-id'>
            REMOVE_ID,
</xsl:when>
<xsl:when test='dryverl:value-map-remove-entry-value'>
            REMOVE_VALUE,
</xsl:when>
<xsl:when test='dryverl:value-map-destroy-entry-id'>
            DESTROY_ID,
</xsl:when>
<xsl:when test='dryverl:value-map-destroy-entry-value'>
            DESTROY_VALUE,
</xsl:when>
</xsl:choose>
<xsl:choose>
<xsl:when test='dryverl:value-map-entry-id-ref'>
            <!--(unsigned long*)-->
	        (<xsl:apply-templates select='dryverl:value-map-entry-id-ref'/>),
</xsl:when>
<xsl:otherwise>
            NULL,
</xsl:otherwise>
</xsl:choose>
<xsl:choose>
<xsl:when test='dryverl:value-map-entry-value-ref'>
            <!--(<xsl:value-of select='ancestor::dryverl:def-bindings/dryverl:def-value-map[@name = $name]/@c-type'/>*)-->
                (<xsl:apply-templates select='dryverl:value-map-entry-value-ref'/>),
</xsl:when>
<xsl:otherwise>
            NULL,
</xsl:otherwise>
</xsl:choose>
            NULL, 0, 0)) { <!-- FIXME: is this really useful? -->
    case 0: <!-- Entry found, added or removed correctly -->
        {
        <xsl:apply-templates select='dryverl:value-map-if-entry-found'/>
        }
        break;
    case 1: <!-- Entry not found -->
        {
        <xsl:apply-templates select='dryverl:value-map-if-entry-not-found'/>
        }
        break;
<xsl:if test='$runtime-checks'>
    default: <!-- Most likely a memory error when trying to allocate
                  an entry -->
        goto memory_error;
        break;
</xsl:if>
    }
</xsl:template>

</xsl:stylesheet>

