<?xml version="1.0"?>
<!--
Copyright (C) 2002, 2003  Scott Lystig Fritchie
Copyright (C) 2006  Romain Lenglet
All rights reserved.
The contents of this file are subject to the Modified BSD License.
See the file "COPYING" for licensing details.
-->

<xsl:stylesheet version='1.0'
    xmlns:xsl='http://www.w3.org/1999/XSL/Transform'
    xmlns:dryverl='http://objectweb.org/dryverl-0.1'>

<xsl:output method='text'/>

<!-- Empty template in case there is no gen-driver element! -->
<xsl:template match='/dryverl:def-bindings[not(dryverl:gen-driver)]'>
</xsl:template>

<xsl:template match='/dryverl:def-bindings[dryverl:gen-driver]'>
<!-- NOTE: this must be synchronized with the same variable definition in the
     .h file stylesheet -->
<xsl:variable name='first-user-command-id' select='2'/>
-module(<xsl:value-of select='dryverl:gen-driver/@erlang-gen-server-module-name'/>).

-behaviour(gen_server).
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

-record(state, {driver_name, port}).

<xsl:apply-templates select='dryverl:gen-driver/dryverl:def-erlang-gen-server-header'/>

<!-- FIXME: For now, no options are allowed (3rd argument) -->
init([DriverDirPath, DriverName, []]) ->
    erl_ddll:start(),
    case erl_ddll:load_driver(DriverDirPath, DriverName) of
        ok -> case open_port({spawn, DriverName}, []) of
                Port when is_port(Port) ->
                    State = #state{
                        driver_name = DriverName,
                        port = Port},
                    {ok, State};
                Error ->
                    {stop, Error}
            end;
        {error, ErrorDescriptor} ->
            {stop, erl_ddll:format_error(ErrorDescriptor)}
    end.

<!-- NOTE: this implementation of handle_call relies on the internal format of
     the From argument, as implemented in the gen and gen_server modules: it is
     a tuple of the client PID and a client-specific reference. -->

<xsl:for-each select='dryverl:def-erlang-to-c-binding[dryverl:encode-output]/dryverl:def-erlang-input'>
handle_call({call, <xsl:value-of select='@function-name'/><xsl:if test='dryverl:def-erlang-arg'>, [<xsl:for-each select='dryverl:def-erlang-arg'><xsl:if test='position() > 1'>, </xsl:if>Arg<xsl:value-of select='@name'/></xsl:for-each>]</xsl:if>}, {FromPid, FromTag}, State) when is_pid(FromPid), is_reference(FromTag) ->
    case call_<xsl:value-of select='@function-name'/>(State#state.port, FromPid, FromTag<xsl:for-each select='dryverl:def-erlang-arg'>, Arg<xsl:value-of select='@name'/></xsl:for-each>) of
        noreply -> {noreply, State};
        {reply, Reply} -> {reply, Reply, State}
    end;
</xsl:for-each>

handle_call({get_driver_option, OptionName}, _From, State) ->
    {reply, get_driver_option(State#state.port, OptionName), State};

handle_call({set_driver_option, OptionName, Value}, _From, State) ->
    set_driver_option(State#state.port, OptionName, Value),
    {reply, ok, State}.

<xsl:for-each select='dryverl:def-erlang-to-c-binding[not(dryverl:encode-output)]/dryverl:def-erlang-input'>
handle_cast({call, <xsl:value-of select='@function-name'/><xsl:if test='dryverl:def-erlang-arg'>, [<xsl:for-each select='dryverl:def-erlang-arg'><xsl:if test='position() > 1'>, </xsl:if>Arg<xsl:value-of select='@name'/></xsl:for-each>]</xsl:if>}, State) ->
    noreply = call_<xsl:value-of select='@function-name'/>(State#state.port<xsl:for-each select='dryverl:def-erlang-arg'>, Arg<xsl:value-of select='@name'/></xsl:for-each>),
    {noreply, State};
</xsl:for-each>

handle_cast({set_driver_option, OptionName, Value}, State) ->
    set_driver_option(State#state.port, OptionName, Value),
    {noreply, State}.


<!-- Undecode and forward the return term received from the driver, if the
     driver has the async option set to true, or if byref binaries had to be
     passed by reference either in input or output -->

<xsl:for-each select='dryverl:def-erlang-to-c-binding/dryverl:def-erlang-input'>
<xsl:variable name='command-id' select='position() + $first-user-command-id - 1'/>
<xsl:variable name='has-output' select='boolean(../dryverl:encode-output)'/>
<xsl:variable name='has-output-byref-bins' select='boolean(../dryverl:encode-output/descendant::dryverl:encode-output-byref-binary)'/>
<xsl:if test='$has-output'>
handle_info({drvreply, Port, <xsl:value-of select='$command-id'/>,
        EncOutputMainTerm<xsl:if test='$has-output-byref-bins'>, OutputByrefBinaryList</xsl:if>}, State) when Port == State#state.port,
        is_binary(EncOutputMainTerm)<xsl:if test='$has-output-byref-bins'>, is_list(OutputByrefBinaryList)</xsl:if> ->
    {FromPid, FromTag, OutputMainTerm} = binary_to_term(EncOutputMainTerm),
    Reply = decode_output_<xsl:value-of select='@function-name'/>(OutputMainTerm<xsl:if test='$has-output-byref-bins'>, OutputByrefBinaryList</xsl:if>),
    <!-- NOTE: this is equivalent to calling
         gen_server:reply({FromPid, FromTag}, Reply): -->
    FromPid ! {FromTag, Reply},
    {noreply, State};
</xsl:if>
</xsl:for-each>
handle_info({drvdebug, Port, Term}, State) when Port == State#state.port ->
    <!-- FIXME: use sasl, etc. -->
    io:format("~w~n", [Term]),
    {noreply, State}.

terminate(_Reason, State) ->
    port_close(State#state.port),
    erl_ddll:unload_driver(State#state.driver_name),
    ok.

code_change(_OldVsn, State, _Extra) ->
    {ok, State}. <!-- FIXME: do nothing for now -->

<xsl:for-each select='dryverl:def-erlang-to-c-binding/dryverl:def-erlang-input'>
<xsl:variable name='command-id' select='position() + $first-user-command-id - 1'/>
<xsl:variable name='has-input-main-term' select='boolean(../dryverl:encode-input/dryverl:encode-input-main-term)'/>
<xsl:variable name='has-input-byref-bins' select='boolean(../dryverl:encode-input/dryverl:encode-input-byref-binary-list)'/>
<xsl:variable name='has-output-byref-bins' select='boolean(../dryverl:encode-output/descendant::dryverl:encode-output-byref-binary)'/>
<xsl:variable name='has-output' select='boolean(../dryverl:encode-output)'/>

<xsl:if test='$has-output'>
%% @spec decode_output_<xsl:value-of select='@function-name'/>(__OutputMainTerm<xsl:if test='$has-output-byref-bins'>, __OutputByrefBinaryList</xsl:if>) -> __Reply
%% __OutputMainTerm = term()<xsl:if test='$has-output-byref-bins'>
%% __OutputByrefBinaryList = [binary()]</xsl:if>
%% __Reply = <xsl:value-of select='../dryverl:def-erlang-output/dryverl:def-erlang-return/@type-doc'/>
<xsl:for-each select='../dryverl:def-erlang-output/dryverl:def-erlang-type-doc'>
%% <xsl:apply-templates/>
</xsl:for-each>
decode_output_<xsl:value-of select='@function-name'/>(
        __OutputMainTerm<xsl:if test='$has-output-byref-bins'>, __ByrefBinaryList</xsl:if>) ->
    <xsl:apply-templates select='../dryverl:decode-output/dryverl:create-output-term'/>.
</xsl:if>

%% @spec call_<xsl:value-of select='@function-name'/>(__Port<xsl:if test='$has-output'>, __FromPid, __FromTag</xsl:if><xsl:for-each select='dryverl:def-erlang-arg'>, <xsl:value-of select='@name'/></xsl:for-each>) ->
%%      noreply<xsl:if test='$has-output'> | {reply, __Reply}</xsl:if>
%% __Port = pid()<xsl:if test='$has-output'>
%% __FromPid = pid()
%% __FromTag = reference()</xsl:if><xsl:for-each select='dryverl:def-erlang-arg'>
%% <xsl:value-of select='@name'/> = <xsl:value-of select='@type-doc'/>
</xsl:for-each>
<xsl:for-each select='dryverl:def-erlang-type-doc'>
%% <xsl:apply-templates/>
</xsl:for-each>
<xsl:if test='$has-output'>
%% __Reply = <xsl:value-of select='../dryverl:def-erlang-output/dryverl:def-erlang-return/@type-doc'/>
<xsl:for-each select='../dryverl:def-erlang-output/dryverl:def-erlang-type-doc'>
%% <xsl:apply-templates/>
</xsl:for-each>
</xsl:if>
call_<xsl:value-of select='@function-name'/>(__Port<xsl:if test='$has-output'>, __FromPid, __FromTag</xsl:if><xsl:for-each select='dryverl:def-erlang-arg'>, __Arg<xsl:value-of select='@name'/></xsl:for-each>) ->

<xsl:if test='$has-input-main-term'>
    __InputMainTerm = (<xsl:apply-templates select='../dryverl:encode-input/dryverl:encode-input-main-term'/>),
</xsl:if>
<xsl:if test='$has-input-byref-bins'>
    __ByrefBinaryList = (<xsl:value-of select='../dryverl:encode-input/dryverl:encode-input-byref-binary-list'/>),
    case __ByrefBinaryList of
        [] ->
</xsl:if>
        <!-- If no byref binaries have to be sent by reference, port_call is
             used instead of port_command -->
        <!-- And if no byref binaries are sent in output, and the driver
             'async' option is set to false at runtime, we can also directly
             return the result directly in the port_call call -->
<xsl:if test='$has-output'>
    case
</xsl:if>
        erlang:port_call(__Port, <xsl:value-of select='$command-id'/>,
<xsl:if test='$has-output'>
        {__FromPid, __FromTag,
</xsl:if>
<xsl:choose>
<xsl:when test='$has-input-main-term'>
        __InputMainTerm
</xsl:when>
<xsl:otherwise>
        []
</xsl:otherwise>
</xsl:choose>
<xsl:if test='$has-output'>
        }
</xsl:if>
        )
<xsl:if test='$has-output'>
        of 
        noreply -> noreply;
        {__FromPid, __FromTag, __OutputMainTerm} ->
            __Reply = decode_output_<xsl:value-of select='@function-name'/>(
                __OutputMainTerm<xsl:if test='$has-output-byref-bins'>, []</xsl:if>),
            {reply, __Reply}
    end
</xsl:if>
<xsl:if test='$has-input-byref-bins'>
        ; _ ->
        <!-- If byref binaries have to be sent by reference, port_command is
             always used -->

        __EncodedMainBinary = term_to_binary({<xsl:value-of select='$command-id'/>
<xsl:if test='$has-output'>
            , __FromPid, __FromTag
</xsl:if>
<xsl:if test='$has-input-main-term'>
            , __InputMainTerm
</xsl:if>
            }),
        <!-- Explicitly send the size of every byref binary: this is necessary
             because all the binaries may be packed (i.e. grouped into a lower
             number of binaries) by port_command, and we want to keep
             boundaries between binaries -->
        __EncodedByrefBinariesSizesList = term_to_binary(
            lists:map(fun(B) -> true = is_binary(B), size(B) end,
                      __ByrefBinaryList)),
        <!-- Prepend a header with the sizes of the encoded main term, and of
             the encoded sizes list, to make parsing possible in case of
             packing -->
        __EncodedHeaderBinary = term_to_binary({size(__EncodedMainBinary),
            size(__EncodedByrefBinariesSizesList)}),
        port_command(__Port, [__EncodedHeaderBinary, __EncodedMainBinary,
            __EncodedByrefBinariesSizesList | __ByrefBinaryList]),
        noreply
    end
</xsl:if>
    .
</xsl:for-each>

%% @spec get_driver_option(Port, OptionName) -> Value
%% Port = port()
%% OptionName = optionName()
%% Value = term()
get_driver_option(Port, OptionName) ->
    erlang:port_call(Port, 0, option_name_to_integer(OptionName)).
<!-- NOTE: the second argument to port_call/3 is the same value as that of the
     ..._COMMAND_GET_OPTION constant in the driver's .h file -->

%% @spec set_driver_option(Port, OptionName, Value) -> ok
%% Port = port()
%% OptionName = optionName()
%% Value = term()
set_driver_option(Port, OptionName, Value) ->
    erlang:port_call(Port, 1,
        {option_name_to_integer(OptionName), Value}).
<!-- NOTE: the second argument to port_call/3 is the same value as that of the
     ..._COMMAND_SET_OPTION constant in the driver's .h file -->

%% @spec option_name_to_integer(Option) -> integer()
%% Option = optionName()
%% optionName() = async
option_name_to_integer(async) -> 0.
<!-- NOTE: the integer values are the same as those of the ..._OPTION_...
     constants in the driver's .h file -->

</xsl:template>

<!-- Templates for elements that can be used in <encode-input-main-term/>
     and <encode-input-byref-binary-list/> -->

<xsl:template match='dryverl:erlang-arg'>
<xsl:choose>
<xsl:when test='not(ancestor::dryverl:def-erlang-to-c-binding/dryverl:def-erlang-input/dryverl:def-erlang-arg/@name = @name)'>
<xsl:message terminate="yes">
Access to an undefined Erlang argument "<xsl:value-of select='@name'/>".
</xsl:message>
</xsl:when>
<xsl:otherwise>
__Arg<xsl:value-of select='@name'/>
</xsl:otherwise>
</xsl:choose>
</xsl:template>

<!-- Templates for elements that can be used in <create-output-term/>
     elements -->

<xsl:template match='dryverl:erl-output-main-term'>
(__OutputMainTerm)
</xsl:template>

<xsl:template match='dryverl:erl-output-byref-binary-list'>
<xsl:choose>
<xsl:when test='ancestor::dryverl:def-erlang-to-c-binding/dryverl:encode-output/descendant::dryverl:encode-output-byref-binary'>
(__ByrefBinaryList)
</xsl:when>
<xsl:otherwise>
([])
</xsl:otherwise>
</xsl:choose>
</xsl:template>

</xsl:stylesheet>

